/*
 * Copyright (c) 2002 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: driver_blur.c,v 1.2 2002/05/24 01:58:20 zinnia Exp $
 */

#include "flip.h"
#include "SDL_rotozoom.h"

/*
DRIVERSPEC blur
*/

typedef struct {
  Uint32* r;
  Uint32* g;
  Uint32* b;
  /*  Uint32* a;*/
} blur_context;

#if 0
SDL_Surface* blurSurface(SDL_Surface* src, blur_context* dummy, int f)
{
  SDL_Surface* tmp;
  int x, y; Uint8* p; int l, factor, count, mod;
  Uint8* offset; Uint8* q;
  Uint32 v;
  Uint32 vr, vg, vb;
  int d;

  SDL_Color color;

  factor = ABS(f) | 1;

  color.r = color.g = color.b = 0;

  tmp = SDL_ConvertSurface(src, src->format, src->flags);
  SDL_FillRect(tmp, NULL, src->format->colorkey);

  for(y = 0; y < tmp->h; y++){
    p = src->pixels + y * src->pitch;
    q = tmp->pixels + y * tmp->pitch;
    for(x = 0; x < src->w; x++){
      vr = vg = vb = 0; count = 0;
      for(l = 0; l < factor / 2; l++) {
	offset = p;
	/* minus */
	d = l * src->format->BytesPerPixel;
	if(x - l >= 0) offset -= d;
	else offset += d;
	d = l * src->pitch;
	if(y - l >= 0) offset -= d;
	else offset += d;

	v = (*((Uint32*)offset));
	vr += EXTRACT_PIXEL(tmp, v, R);
	vg += EXTRACT_PIXEL(tmp, v, G);
	vb += EXTRACT_PIXEL(tmp, v, B);
	count++;

	offset = p;
	/* plus */
	d = l * src->format->BytesPerPixel;
	if(x + l < src->w) offset += d;
	else offset -= d;

	d = l * src->pitch;
	if(y + l < src->h) offset += d;
	else offset += d;

	v = (*((Uint32*)offset));
	vr += EXTRACT_PIXEL(tmp, v, R);
	vg += EXTRACT_PIXEL(tmp, v, G);
	vb += EXTRACT_PIXEL(tmp, v, B);
	count++;
      }
      if(count > 0) {
	vr /= count;
	vg /= count;
	vb /= count;
      }

      *((Uint32*)q) = COMPOSE_PIXEL(tmp,vr,vg,vb);
      p += src->format->BytesPerPixel;
      q += tmp->format->BytesPerPixel;
    }
  }
  return tmp;
}

#else
SDL_Surface* blurSurface(SDL_Surface* src, blur_context* bb, int factor)
{
  int x, y, dx, dy, offset, r, g, b, i, j, f, offset2, modulo;
  Uint32 v;
  Uint8* p;
  SDL_Surface* tmp;

  f = ABS(factor) | 1;
  modulo = 0;
  if(f > 7){
    modulo = f - 7;
    f = 7;
  }
  memset(bb->r, 0, sizeof(Uint32) * src->w * src->h);
  memset(bb->g, 0, sizeof(Uint32) * src->w * src->h);
  memset(bb->b, 0, sizeof(Uint32) * src->w * src->h);

  offset = 0;
  for(y = 0; y < src->h; y++){
    p = src->pixels + y * src->pitch;
    for(x = 0; x < src->w; x++){
      v = *((Uint32*)p);
      r = EXTRACT_PIXEL(src, v, R);
      g = EXTRACT_PIXEL(src, v, G);
      b = EXTRACT_PIXEL(src, v, B);

      for(dy = -f/2; dy <= f/2; dy++){
	if(y + dy < 0 || y + dy >= src->h) continue;
	offset2 = dy * src->w;
	for(dx = -f/2; dx <= f/2; dx++){
	  if(x + dx < 0 || x + dx >= src->w) continue;
	  bb->r[offset + offset2 + dx] += r;
	  bb->g[offset + offset2 + dx] += g;
	  bb->b[offset + offset2 + dx] += b;
	}
      }
      bb->r[offset] += 512 * modulo;
      bb->g[offset] += 512 * modulo;
      bb->b[offset] += 512 * modulo;
      offset++;
      p += src->format->BytesPerPixel;
    }
  }
  tmp = SDL_ConvertSurface(src, src->format, src->flags);
  p = tmp->pixels;
  offset = 0;
  for(y = 0; y < src->h; y++){
    p = tmp->pixels + tmp->pitch * y;
    for(x = 0; x < src->w; x++){
      v = f * f;
#define MIN(a,b) ((a) <= (b) ? (a) : (b))

      r = MIN(bb->r[offset] / v, 255);
      g = MIN(bb->g[offset] / v, 255);
      b = MIN(bb->b[offset] / v, 255);
      v = COMPOSE_PIXEL(tmp, r, g, b);
      *((Uint32*)p) = v;
      offset++;
      p += tmp->format->BytesPerPixel;
    }
  }
  return tmp;

}
#endif

#define BLUR_FACTOR 31

void blur_flip(SDL_Surface* screen, SDL_Surface* bmp)
{
  int i;
  SDL_Surface* src;
  SDL_Surface* tmp = NULL;
  blur_context b;
  SDL_Rect dst;

  int size;

  src = SDL_ConvertSurface(screen, screen->format, screen->flags);
  size = sizeof(Uint32) * src->w * src->h;

  b.r = (Uint32*) malloc(size);
  b.g = (Uint32*) malloc(size);
  b.b = (Uint32*) malloc(size);
  
  for(i = 3; i < BLUR_FACTOR; i += 2){
    fprintf(stderr, "%d\n", i);
    tmp = blurSurface(src, &b, i);
    dst.x = dst.y = 0;
    dst.w = src->w; dst.h = src->h;
    SDL_BlitSurface(tmp, NULL, screen, &dst);
    SDL_UpdateRect(screen, 0, 0, 0, 0);
    SDL_Delay(20);
    SDL_FreeSurface(tmp);
  }
  for(i = BLUR_FACTOR; i >= 1; i -= 2){
    fprintf(stderr, "%d\n", i);
    tmp = blurSurface(bmp, &b, i);
    dst.x = dst.y = 0;
    dst.w = bmp->w; dst.h = bmp->h;
    SDL_BlitSurface(tmp, NULL, screen, &dst);
    SDL_UpdateRect(screen, 0, 0, 0, 0);
    SDL_Delay(20);
    SDL_FreeSurface(tmp);
  }
}
